"""Module for stereo camera rectification."""

from dataclasses import dataclass
from typing import TypeAlias

import cv2
import numpy as np

import mynd.schemas as schemas

from mynd.image import Image
from mynd.utils.containers import Pair
from mynd.utils.log import logger

from .pixel_map import (
    PixelMap,
    compute_pixel_map,
    remap_image_pixels,
    invert_pixel_map,
)


PixelMapPair: TypeAlias = Pair[PixelMap]


def rectify_stereo_sensors(sensors: schemas.StereoSensorPairSchema) -> None:
    """Computes stereo rectified sensors and updates the rectified sensor
    components."""
    rectified_master: schemas.RectifiedSensorSchema
    rectified_slave: schemas.RectifiedSensorSchema
    rectified_master, rectified_slave = compute_stereo_rectified_sensors(
        sensors.master.sensor,
        sensors.slave.sensor,
    )

    # Update the rectified components of the sensors
    sensors.master.sensor.rectified = rectified_master
    sensors.slave.sensor.rectified = rectified_slave


def compute_stereo_rectified_sensors(
    master_sensor: schemas.SensorSchema,
    slave_sensor: schemas.SensorSchema,
) -> tuple[schemas.RectifiedSensorSchema, schemas.RectifiedSensorSchema]:
    """Computes the camera calibrations for a pair of rectified cameras."""

    assert isinstance(master_sensor, schemas.SensorSchema), "invalid master sensor type"
    assert isinstance(slave_sensor, schemas.SensorSchema), "invalid master sensor type"

    transforms: Transforms = _compute_stereo_rectifying_transforms(
        master_sensor, slave_sensor
    )

    resolution_master: tuple[int, int] = (
        master_sensor.calibration.width,
        master_sensor.calibration.height,
    )
    resolution_slave: tuple[int, int] = (
        slave_sensor.calibration.width,
        slave_sensor.calibration.height,
    )

    desired_resolution: tuple[int, int] = resolution_master

    affine_transform: np.ndarray = _compute_common_affine_transform(
        master_sensor.calibration.projection_matrix,  # intrinsic 1
        slave_sensor.calibration.projection_matrix,  # intrinsic 2
        transforms.homographies.first,  # homography 1
        transforms.homographies.second,  # homography 2
        resolution_master,  # resolution 1
        resolution_slave,  # resolution 2
        master_sensor.calibration.distortion_vector,  # distortion 1
        slave_sensor.calibration.distortion_vector,  # distortion 2
        desired_resolution,  # desired resolution - (width, height)
    )

    # Group all the transformations applied after rectification
    # These would be needed for 3D reconstrunction
    new_first_camera: np.ndarray = (
        affine_transform.dot(transforms.homographies.first)
        .dot(master_sensor.calibration.projection_matrix)
        .dot(master_sensor.rotation_matrix)
        .dot(transforms.rotation.T)
    )

    new_second_camera: np.ndarray = (
        affine_transform.dot(transforms.homographies.second)
        .dot(slave_sensor.calibration.projection_matrix)
        .dot(slave_sensor.rotation_matrix)
        .dot(transforms.rotation.T)
    )

    updated_camera_matrices: Pair[np.ndarray] = Pair(
        first=new_first_camera,
        second=new_second_camera,
    )

    master_calibration: schemas.CalibrationSchema = _create_camera_calibration(
        width=desired_resolution[0],
        height=desired_resolution[1],
        projection_matrix=updated_camera_matrices.first,
    )

    slave_calibration: schemas.CalibrationSchema = _create_camera_calibration(
        width=desired_resolution[0],
        height=desired_resolution[1],
        projection_matrix=updated_camera_matrices.second,
    )

    rectified_master_sensor: schemas.RectifiedSensorSchema = (
        schemas.RectifiedSensorSchema(
            label=f"{master_sensor.label}_rectified",
            width=desired_resolution[0],
            height=desired_resolution[1],
            location=np.zeros(3).tolist(),
            rotation=transforms.rotation.tolist(),
            calibration=master_calibration,
        )
    )

    baseline: float = np.linalg.norm(
        slave_sensor.location_vector - master_sensor.location_vector
    )

    rectified_slave_sensor: schemas.RectifiedSensorSchema = (
        schemas.RectifiedSensorSchema(
            label=f"{slave_sensor.label}_rectified",
            width=desired_resolution[0],
            height=desired_resolution[1],
            location=np.array([baseline, 0.0, 0.0]).tolist(),
            rotation=transforms.rotation.dot(slave_sensor.rotation_matrix.T).tolist(),
            calibration=slave_calibration,
        )
    )

    return rectified_master_sensor, rectified_slave_sensor


def compute_sensor_rectification_map(sensor: schemas.SensorSchema) -> PixelMap:
    """Computes the pixel map between an unrectified and rectified sensor."""

    assert sensor.is_rectified(), "sensor is not rectified"
    assert isinstance(sensor, schemas.SensorSchema), "invalid sensor type"

    target_size: tuple[int, int] = (
        sensor.rectified.calibration.width,
        sensor.rectified.calibration.height,
    )

    pixel_map: PixelMap = compute_pixel_map(
        sensor.calibration.projection_matrix,
        sensor.calibration.distortion_vector,
        sensor.rectified.rotation_matrix,
        sensor.rectified.calibration.projection_matrix,
        target_size,
    )

    return pixel_map


def compute_stereo_rectification_map(
    sensors: schemas.StereoSensorPairSchema,
) -> PixelMapPair:
    """Computes updated camera matrices and pixel maps based on the given stereo calibration
    and rectifying transforms."""
    assert isinstance(sensors, schemas.StereoSensorPairSchema), "invalid sensors type"
    assert sensors.master.is_rectified(), "master sensor is not rectified"
    assert sensors.slave.is_rectified(), "slave sensor is not rectified"

    # Recompute final maps considering fitting transformations too
    master_pixel_map: PixelMap = compute_sensor_rectification_map(sensors.master.sensor)
    slave_pixel_map: PixelMap = compute_sensor_rectification_map(sensors.slave.sensor)

    pixel_maps: Pair[PixelMap] = Pair(
        first=master_pixel_map,
        second=slave_pixel_map,
    )

    return pixel_maps


@dataclass(frozen=True)
class Transforms:
    """Class representing rectification transforms including a common
    rotation and homographies for the two cameras."""

    rotation: np.ndarray  # common rotation for the two cameras
    homographies: Pair[np.ndarray]  # transforms for the two camera


def _compute_stereo_rectifying_transforms(
    master_sensor: schemas.SensorSchema,
    slave_sensor: schemas.SensorSchema,
) -> Transforms:
    """
    Computes the rectifying transforms for a pair of stereo sensor using the
    standard OpenCV algorithm.
    Adopted from: https://github.com/decadenza/SimpleStereo/blob/master/simplestereo/_rigs.py
    """
    assert isinstance(master_sensor, schemas.SensorSchema), "invalid master sensor type"
    assert isinstance(slave_sensor, schemas.SensorSchema), "invalid slave sensor type"

    resolution: tuple[int, int] = (
        master_sensor.calibration.width,
        master_sensor.calibration.height,
    )

    first_rotation, second_rotation, _, _, _, _, _ = cv2.stereoRectify(
        master_sensor.calibration.projection_matrix,  # 3x3 master camera matrix
        master_sensor.calibration.distortion_vector,  # 5x1 master camera distortion
        slave_sensor.calibration.projection_matrix,  # 3x3 slave camera matrix
        slave_sensor.calibration.distortion_vector,  # 5x1 slave camera distortion
        resolution,  # resolution (width, height)
        slave_sensor.rotation_matrix,  # 3x3 rotation matrix from camera 1 to camera 2
        slave_sensor.location_vector,  # 3x1 translation vector from camera 1 to camera 2
        flags=0,
    )

    # OpenCV does not compute the rectifying homography, but a rotation in the object space.
    # R1 = Rnew * Rcam^{-1}
    # To get the homography:
    first_homography: np.ndarray = first_rotation.dot(
        np.linalg.inv(master_sensor.calibration.projection_matrix)
    )
    second_homography: np.ndarray = second_rotation.dot(
        np.linalg.inv(slave_sensor.calibration.projection_matrix)
    )

    homographies: Pair[np.ndarray] = Pair(
        first=first_homography, second=second_homography
    )

    # To get the common orientation, since the first camera has orientation as origin:
    # Rcommon = R1
    # It also can be retrieved from R2, cancelling the rotation of the second camera.
    # Rcommon = R2.dot(np.linalg.inv(rig.R))

    return Transforms(rotation=first_rotation, homographies=homographies)


def _create_camera_calibration(
    width: int,
    height: int,
    projection_matrix: np.ndarray,
    distortion_vector: np.ndarray = np.zeros(5, dtype=np.float32),
) -> schemas.CalibrationSchema:
    """Converts a camera matrix to a calibration."""
    assert projection_matrix.shape == (3, 3), "invalid projection matrix shape"
    assert distortion_vector.shape == (5,), "invalid distortion vector shape"

    assert projection_matrix.dtype in [
        np.float32,
        np.float64,
    ], f"invalid projection dtype: {projection_matrix.dtype}"
    assert distortion_vector.dtype in [
        np.float32,
        np.float64,
    ], f"invalid distortion dtype: {distortion_vector.dtype}"

    fx: float = projection_matrix[0, 0]
    fy: float = projection_matrix[1, 1]
    cx: float = projection_matrix[0, 2]
    cy: float = projection_matrix[1, 2]

    k1: float = distortion_vector[0]
    k2: float = distortion_vector[1]
    p1: float = distortion_vector[2]
    p2: float = distortion_vector[3]
    k3: float = distortion_vector[4]

    return schemas.CalibrationSchema(
        width=width,
        height=height,
        fx=fx,
        fy=fy,
        cx=cx,
        cy=cy,
        k1=k1,
        k2=k2,
        k3=k3,
        p1=p1,
        p2=p2,
    )


def _compute_common_affine_transform(
    first_camera_matrix: np.ndarray,
    second_camera_matrix: np.ndarray,
    first_homography: np.ndarray,
    second_homography: np.ndarray,
    first_dims: tuple[int, int],
    second_dims: tuple[int, int],
    first_distortion: np.ndarray | None = None,
    second_distortion: np.ndarray | None = None,
    desired_dims: tuple[int, int] | None = None,
) -> np.ndarray:
    """
    Compute affine tranformation to fit the rectified images into desidered dimensions.

    After rectification usually the image is no more into the original image bounds.
    One can apply any transformation that do not affect disparity to fit the image into boundaries.
    This function corrects flipped images too.
    The algorithm may fail if one epipole is too close to the image.

    Parameters
    ----------
    first_camera_matrix, second_camera_matrix : numpy.ndarray
        3x3 original camera matrices of intrinsic parameters.
    first_homography, second_homography : numpy.ndarray
        3x3 rectifying transforms.
    first_dims, second_dims : tuple
        Resolution of images as (width, height) tuple.
    first_distortion, second_distortion : numpy.ndarray, optional
        Distortion coefficients in the order followed by OpenCV. If None is passed, zero distortion is
        assumed.
    desired_dims : tuple, optional
        Resolution of destination images as (width, height) tuple (default to the first image
        resolution).

    Returns
    -------
    numpy.ndarray
        3x3 affine transformation to be used both for the first and for the second camera.
    """

    if first_distortion is not None:
        assert first_distortion.shape == (
            5,
        ), f"invalid distortion shape: {first_distortion.shape}"
    if second_distortion is not None:
        assert second_distortion.shape == (
            5,
        ), f"invalid distortion shape: {second_distortion.shape}"

    if desired_dims is None:
        desired_dims = first_dims

    desired_width: int
    desired_height: int
    desired_width, desired_height = desired_dims

    first_corners: ImageCorners = compute_image_corners(
        homography=first_homography,
        camera_matrix=first_camera_matrix,
        dimensions=first_dims,
        distortion=first_distortion,
    )

    second_corners: ImageCorners = compute_image_corners(
        homography=second_homography,
        camera_matrix=second_camera_matrix,
        dimensions=second_dims,
        distortion=second_distortion,
    )

    min_x1: float = first_corners.min[0]
    max_x1: float = first_corners.max[0]

    min_x2: float = second_corners.min[0]
    max_x2: float = second_corners.max[0]

    min_y: float = min(first_corners.min[1], second_corners.min[1])
    max_y: float = max(first_corners.max[1], second_corners.max[1])

    # Flip factor
    flip_x: int = 1
    flip_y: int = 1

    if first_corners.top_left[0] > first_corners.top_right[0]:
        flip_x: int = -1
    if first_corners.top_left[1] > first_corners.bottom_left[1]:
        flip_y: int = -1

    # Scale X (choose common scale X to best fit bigger image between left and right)
    if max_x2 - min_x2 > max_x1 - min_x1:
        scale_x: float = flip_x * desired_width / (max_x2 - min_x2)
    else:
        scale_x: float = flip_x * desired_width / (max_x1 - min_x1)

    # Scale Y (unique not to lose rectification)
    scale_y: float = flip_y * desired_height / (max_y - min_y)

    # Translation X (keep always at left border)
    if flip_x == 1:
        translation_x: float = -min(min_x1, min_x2) * scale_x
    else:
        translation_x: float = -min(max_x1, max_x2) * scale_x

    # Translation Y (keep always at top border)
    if flip_y == 1:
        translation_y: float = -min_y * scale_y
    else:
        translation_y: float = -max_y * scale_y

    # Final affine transformation
    affine: np.ndarray = np.array(
        [[scale_x, 0, translation_x], [0, scale_y, translation_y], [0, 0, 1]]
    )

    return affine


@dataclass(frozen=True)
class ImageCorners:
    """Class representing image corners."""

    top_left: tuple[float, float]  # corners[0,0] = [0,0]
    top_right: tuple[float, float]  # corners[1,0] = [dims[0]-1,0]
    bottom_right: tuple[float, float]  # corners[2,0] = [dims[0]-1,dims[1]-1]
    bottom_left: tuple[float, float]  # corners[3,0] = [0, dims[1]-1]

    @property
    def min(self) -> tuple[float, float]:
        """Returns the minimum x- and y-coordinate of the image corners."""
        minx: float = min(
            self.top_left[0],
            self.top_right[0],
            self.bottom_right[0],
            self.bottom_left[0],
        )
        miny: float = min(
            self.top_left[1],
            self.top_right[1],
            self.bottom_right[1],
            self.bottom_left[1],
        )
        return (minx, miny)

    @property
    def max(self) -> tuple[float, float]:
        """Returns the maximum x- and y-coordinate of the image corners."""
        maxx: float = max(
            self.top_left[0],
            self.top_right[0],
            self.bottom_right[0],
            self.bottom_left[0],
        )
        maxy: float = max(
            self.top_left[1],
            self.top_right[1],
            self.bottom_right[1],
            self.bottom_left[1],
        )
        return (maxx, maxy)


def compute_image_corners(
    homography: np.ndarray,
    camera_matrix: np.ndarray,
    dimensions: tuple[int, int],
    distortion: np.ndarray | None = None,
) -> ImageCorners:
    """Computes updated image corners locations based on a combined homography
    and undistortion transformation."""

    width: int
    height: int
    width, height = dimensions

    if distortion is None:
        distortion: np.ndarray = np.zeros(5)

    # Set image corners in the form requested by cv2.undistortPoints
    corners: np.ndarray = np.zeros((4, 1, 2), dtype=np.float32)

    # Initialize corners in order: top left, top right, bottom right, bottom left
    corners[0, 0] = [0, 0]
    corners[1, 0] = [width - 1, 0]
    corners[2, 0] = [width - 1, height - 1]
    corners[3, 0] = [0, height - 1]

    undistorted_corners: np.ndarray = cv2.undistortPoints(
        corners, camera_matrix, distortion, R=homography.dot(camera_matrix)
    )

    undistorted_corners: list[tuple] = [
        (x, y) for x, y in np.squeeze(undistorted_corners)
    ]

    return ImageCorners(*undistorted_corners)



"""Module for creating plotting figures."""

import plotly.graph_objects as go

from plotly.subplots import make_subplots

from mynd.geometry.transformations import decompose_rigid_transform
from mynd.geometry.transformations import rotation_matrix_to_euler
from mynd.pipelines.registration import RegistrationResult


def create_subplots(
    rows: int,
    cols: int,
    row_heights: list[int] = None,
    column_widths: list[int] = None,
) -> go.Figure:
    """Creates a figure with subplots with placeholder titles."""

    if not row_heights:
        row_heights = [1] * rows
    if not column_widths:
        column_widths = [1] * cols

    subplot_titles: tuple[str] = tuple(f"Plot {index}" for index in range(rows * cols))

    figure: go.Figure = make_subplots(
        rows=rows,
        cols=cols,
        row_heights=row_heights,
        column_widths=column_widths,
        subplot_titles=subplot_titles,
    )

    return figure


def trace_registration_result(
    result: RegistrationResult,
    name: str,
    legendgroup: int,
    color: str = "blue",
) -> dict[str, go.Trace]:
    """Creates graph objects for a registration result. The fitness, error, and correspondence count
    are plotted in addition to the transformation components."""

    scale, rotation, translation = decompose_rigid_transform(result.transformation)
    rotz, roty, rotx = rotation_matrix_to_euler(rotation, degrees=True)

    traces: dict = dict()

    traces["fitness"] = go.Bar(
        name=name,
        x=["Fitness"],
        y=[result.fitness],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=True,
    )

    traces["rmse"] = go.Bar(
        name=name,
        x=["RMSE"],
        y=[result.inlier_rmse],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=False,
    )

    traces["correspondences"] = go.Bar(
        name=name,
        x=["Correspondences"],
        y=[len(result.correspondence_set)],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=False,
    )

    traces["scale"] = go.Bar(
        name=name,
        x=["Scale"],
        y=[scale],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=False,
    )

    traces["rotation"] = go.Bar(
        name=name,
        x=["Rz", "Ry", "Rx"],
        y=[rotz, roty, rotx],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=False,
    )

    traces["translation"] = go.Bar(
        name=name,
        x=["Tx", "Ty", "Tz"],
        y=[translation[0], translation[1], translation[2]],
        marker_color=color,
        hoverinfo="x+y",
        legendgroup=legendgroup,
        showlegend=False,
    )

    return traces



"""Package for stereo vision, including stereo matching, range and normal map
estimation, and geometric image transformations."""

from .hitnet import create_hitnet_matcher

# NOTE: Consider moving to image module
from .pixel_map import (
    PixelMap,
    compute_pixel_map,
    invert_pixel_map,
    remap_image_pixels,
)

from .range_map import (
    compute_ranges_from_disparities,
    compute_points_from_ranges,
    compute_normals_from_ranges,
    fill_range_map_dilation,
)

from .stereo_geometry import (
    StereoGeometry,
    compute_stereo_geometry,
    compute_stereo_range_maps,
    distort_stereo_geometry,
)

from .stereo_matcher import StereoMatcher

from .stereo_rectification import (
    compute_sensor_rectification_map,
    compute_stereo_rectification_map,
    compute_stereo_rectified_sensors,
    rectify_stereo_sensors,
)


__all__ = [
    "create_hitnet_matcher",
    # ...
    "PixelMap",
    "compute_pixel_map",
    "invert_pixel_map",
    "remap_image_pixels",
    # ...
    "PointCloud",
    "PointCloudLoader",
    "PointCloudProcessor",
    # ...
    "downsample_point_cloud",
    "estimate_point_cloud_normals",
    "create_downsampler",
    "create_normal_estimator",
    # ...
    "compute_range_from_disparity",
    "compute_points_from_range",
    "compute_normals_from_range",
    "fill_range_map_dilation",
    # ...
    "StereoGeometry",
    "compute_stereo_geometry",
    "compute_stereo_range_maps",
    "distort_stereo_geometry",
    # ...
    "StereoMatcher",
    # ...
    "compute_sensor_rectification_map",
    "compute_stereo_rectification_map",
    "compute_stereo_rectified_sensors",
    "rectify_stereo_sensors",
]



"""Module for literal functionality."""

import ast


Primitive = int | str | float | bool | None


def literal_primitive(literal: str) -> Primitive | None:
    """Evaluate a string as a dtype."""
    try:
        evaluated: Primitive = ast.literal_eval(literal)
        return evaluated
    except ValueError:
        return None



"""Module for batch registration."""

from collections.abc import Callable, Hashable, Sequence
from dataclasses import dataclass, field
from typing import Generic, TypeAlias, TypeVar

import tqdm

from mynd.geometry.point_cloud import PointCloud, PointCloudLoader

from .pipeline import RegistrationPipeline, apply_registration_pipeline
from .types import RegistrationResult


Key: TypeVar = TypeVar("Key", bound=Hashable)


@dataclass(frozen=True)
class RegistrationBatch(Generic[Key]):
    """Class representing a registration batch."""

    @dataclass(frozen=True)
    class PairIndex:
        """Class representing a pair-wise registration index."""

        target: Key
        source: Key

    @dataclass(frozen=True)
    class PairResult:
        """Class representing a pair-wise registration result."""

        target: Key
        source: Key
        result: RegistrationResult

    @dataclass
    class Result:
        """Class representing a batch registration result."""

        target: Key
        sources: dict[Key, RegistrationResult]

    indices: list[PairIndex] = field(default_factory=list)
    loaders: dict[Key, PointCloudLoader] = field(default_factory=dict)

    def keys(self) -> list[Key]:
        """Returns the keys in the registration batch."""
        return self.loaders.keys()

    def get(self, key: Key) -> PointCloudLoader | None:
        """Returns the point cloud loader or none."""
        return self.loaders.get(key)


def generate_indices_one_way(
    target: Key, sources: Sequence[Key]
) -> list[RegistrationBatch.PairIndex]:
    """Generate a collection of registration indices with a single target."""
    return [
        RegistrationBatch.PairIndex(target, source)
        for source in sources
        if source != target
    ]


def generate_indices_cascade(items: Sequence[Key]) -> list[RegistrationBatch.PairIndex]:
    """Generates a list of cascaded multi-source indices."""

    items: list[Key] = list(items)

    indices: list[RegistrationBatch.PairIndex] = list()
    for index, target in enumerate(items[:-1]):
        sources: list[Key] = items[index + 1 :]

        for source in sources:
            indices.append(RegistrationBatch.PairIndex(target=target, source=source))

    return indices


Batch: TypeAlias = RegistrationBatch
Pipeline: TypeAlias = RegistrationPipeline


Callback: TypeAlias = Callable[[Key, Key, RegistrationResult], None]


def register_batch(
    batch: Batch,
    pipeline: Pipeline,
    # indices: list[Batch.PairIndex],
    callback: Callback | None = None,
) -> list[Batch.PairResult]:
    """Registers a batch of point clouds with the given pipeline."""

    results: list[Batch.PairResult] = list()
    for index in tqdm.tqdm(batch.indices, desc="registering batch..."):
        target_loader: PointCloudLoader = batch.loaders.get(index.target)
        target_cloud: PointCloud = target_loader()

        source_loader: PointCloudLoader = batch.loaders.get(index.source)
        source_cloud: PointCloud = source_loader()

        assert isinstance(target_cloud, PointCloud), "invalid target point cloud"
        assert isinstance(source_cloud, PointCloud), "invalid source point cloud"

        result: RegistrationResult = apply_registration_pipeline(
            pipeline,
            target=target_cloud,
            source=source_cloud,
        )

        if callback is not None:
            callback(index.target, index.source, result)

        pairwise: Batch.PairResult = Batch.PairResult(
            target=index.target, source=index.source, result=result
        )

        results.append(pairwise)

    return results



"""Package for working with point cloud data, including types, processors and I/O functionality."""

from .point_cloud_io import (
    read_point_cloud,
    create_point_cloud_loader,
)

from .point_cloud_processors import (
    downsample_point_cloud,
    estimate_point_cloud_normals,
    create_downsampler,
    create_normal_estimator,
)

from .point_cloud_types import (
    PointCloud,
    PointCloudLoader,
    PointCloudProcessor,
)

__all__ = [
    "PointCloudLoader",
    "read_point_cloud",
    "create_point_cloud_loader",
]



"""Module for various point cloud visualization methods."""

import copy

import numpy as np
import open3d.visualization as vis

from mynd.geometry.point_cloud import PointCloud, PointCloudLoader


def create_visualizer(title: str = "Visualization") -> vis.Visualizer:
    """Creates a visualizer that can be used for drawing geometries."""

    visualizer = vis.Visualizer()
    visualizer.create_window(window_name=title)

    return visualizer


def visualize_registration(
    source: PointCloud,
    target: PointCloud,
    transformation: np.ndarray,
    source_color: list | None = None,
    target_color: list | None = None,
    title: str = "",
    window_width: int = 1024,
    window_height: int = 768,
) -> None:
    """Visualizes a pairwise registration result between two point clouds."""
    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)

    if source_color:
        source_temp.paint_uniform_color(source_color)
    if target_color:
        target_temp.paint_uniform_color(target_color)

    source_temp.transform(transformation)

    vis.draw_geometries(
        geometry_list=[source_temp, target_temp],
        window_name=title,
        width=window_width,
        height=window_height,
    )


def visualize_registration_batch(
    storage: dict[int, dict],
    loaders: dict[int, PointCloudLoader],
    source_color: list = [0.60, 0.20, 0.20],
    target_color: list = [0.20, 0.20, 0.60],
) -> None:
    """Visualizes a batch of pairwise point cloud registrations."""
    for source, registrations in storage.items():
        source_cloud: PointCloud = loaders[source]().unwrap()

        for target, result in registrations.items():
            target_cloud: PointCloud = loaders[target]().unwrap()

            visualize_registration(
                source=source_cloud,
                target=target_cloud,
                transformation=result.transformation,
                source_color=source_color,
                target_color=target_color,
                title=f"Source: {source}, target: {target}",
            )



"""Module for registration utility functionality."""

import numpy as np

from mynd.geometry.transformations import decompose_rigid_transform
from mynd.geometry.transformations import rotation_matrix_to_euler
from mynd.utils.log import logger

from .types import RegistrationResult


def log_registration_result(result: RegistrationResult) -> None:
    """Logs a registration result."""

    scale, rotation, translation = decompose_rigid_transform(result.transformation)

    rotz, roty, rotx = rotation_matrix_to_euler(rotation, degrees=True)

    correspondence_count: int = len(result.correspondence_set)

    np.set_printoptions(precision=3)

    logger.info(f"Corresp.:     {correspondence_count}")
    logger.info(f"Fitness:      {result.fitness:.5f}")
    logger.info(f"Inlier RMSE:  {result.inlier_rmse:.5f}")
    logger.info(f"Scale:        {scale:.3f}")
    logger.info(f"Translation:  {translation}")
    logger.info(f"Rot. ZYX:     {rotz:.2f}, {roty:.2f}, {rotx:.2f}")



"""Package for the mynd library"""

import mynd.config as config
import mynd.database as database
import mynd.geometry as geometry
import mynd.image as image
import mynd.records as records
import mynd.registration as registration
import mynd.schemas as schemas
import mynd.tasks as tasks
import mynd.utils as utils
import mynd.visualization as visualization

__all__ = [
    "config",
    "database",
    "geometry",
    "image",
    "records",
    "registration",
    "schemas",
    "tasks",
    "utils",
    "visualization",
]

__version__ = "0.1.1"



"""Module for image visualization functionality."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import NamedTuple, Optional

import cv2
import numpy as np

from mynd.image import Image, PixelFormat
from mynd.image import convert_image_to_rgb, normalize_image, apply_color_map

from mynd.geometry.stereo import StereoGeometry
from mynd.geometry.stereo import PixelMap, remap_image_pixels

from mynd.utils.containers import Pair
from mynd.utils.key_codes import KeyCode


class WindowHandle(NamedTuple):
    """Class representing a window handle."""

    name: str
    width: int
    height: int


class TrackbarData(NamedTuple):
    """Class representing trackbar data."""

    name: str
    lower: int | float
    upper: int | float
    callback: Callable[[int | float], None]


def create_window(
    window_name: str = "Window",
    width: int = 800,
    height: int = 1200,
    track_bars: Optional[list[TrackbarData]] = None,
) -> WindowHandle:
    """Creates a window with optional track bars."""

    cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(window_name, width, height)

    if track_bars:
        for track_bar in track_bars:
            cv2.createTrackbar(
                track_bar.name,
                window_name,
                track_bar.lower,
                track_bar.upper,
                track_bar.callback,
            )

    return WindowHandle(name=window_name, width=width, height=height)


def destroy_window(window: WindowHandle) -> None:
    """Destroys a window."""
    cv2.destroyWindow(window.name)


def destroy_all_windows() -> None:
    """Destroys all the windows."""
    cv2.destroyAllWindows()


def render_image(window: WindowHandle, image: Image | np.ndarray) -> None:
    """Renders an array of values into an image."""
    if isinstance(image, Image):
        values: np.ndarray = image.to_array()
    else:
        values: np.ndarray = image

    cv2.imshow(window.name, values)


def wait_key_input(wait: int) -> KeyCode:
    """Waits for a keyboard input."""
    key: int = cv2.waitKey()
    try:
        key_code: KeyCode = KeyCode(key)
    except ValueError:
        key_code: KeyCode = KeyCode.NULL
    return key_code


@dataclass
class StereoWindows:
    """Class representing a collection of windows for rendering stereo data."""

    color_left: WindowHandle
    color_right: WindowHandle

    range_left: WindowHandle
    range_right: WindowHandle


def create_stereo_windows() -> StereoWindows:
    """Creates a collection of windows for rendering stereo geometries."""
    return StereoWindows(
        color_left=create_window("color_left", 680, 512),
        color_right=create_window("color_right", 680, 512),
        range_left=create_window("range_left", 680, 512),
        range_right=create_window("range_right", 680, 512),
    )


def distort_range_maps(ranges: Pair[Image], pixel_maps: Pair[PixelMap]) -> Pair[Image]:
    """Distorts a pair of range maps."""

    first: Image = remap_image_pixels(image=ranges.first, pixel_map=pixel_maps.first)
    second: Image = remap_image_pixels(image=ranges.second, pixel_map=pixel_maps.second)

    return Pair(first, second)


def render_stereo_geometry(
    windows: StereoWindows,
    geometry: StereoGeometry,
    distort: bool,
) -> None:
    """Render a stereo geometry into a collection of windows."""

    rectification = geometry.rectification

    normalized_ranges: Pair[Image] = Pair(
        first=normalize_image(
            geometry.range_maps.first, lower=0.0, upper=8.0, flip=True
        ),
        second=normalize_image(
            geometry.range_maps.second, lower=0.0, upper=8.0, flip=True
        ),
    )

    colored_ranges: Pair[Image] = Pair(
        first=apply_color_map(normalized_ranges.first),
        second=apply_color_map(normalized_ranges.second),
    )

    images: Pair[Image] = geometry.rectified_images

    if distort:
        images: Pair[Image] = geometry.raw_images
        colored_ranges: Pair[Image] = distort_range_maps(
            colored_ranges, rectification.inverse_pixel_maps
        )

    # Render images
    render_image(windows.color_left, images.first)
    render_image(windows.color_right, images.second)

    # Render range maps
    render_image(windows.range_left, colored_ranges.first)
    render_image(windows.range_right, colored_ranges.second)


def create_stereo_geometry_color_image(
    colors: Pair[Image],
    ranges: Pair[Image],
) -> Image:
    """Creates image tiles for a stereo geometry. Useful for visualizing the
    geometry as RGB."""

    colors: Pair[Image] = Pair(
        convert_image_to_rgb(colors.first),
        convert_image_to_rgb(colors.second),
    )

    normalized_ranges: Pair[Image] = Pair(
        first=normalize_image(ranges.first, lower=0.0, upper=8.0, flip=True),
        second=normalize_image(ranges.second, lower=0.0, upper=8.0, flip=True),
    )

    colored_ranges: Pair[Image] = Pair(
        first=apply_color_map(normalized_ranges.first),
        second=apply_color_map(normalized_ranges.second),
    )

    stacked_colors: np.ndarray = np.hstack(
        (colors.first.to_array(), colors.second.to_array())
    )
    stacked_ranges: np.ndarray = np.hstack(
        (colored_ranges.first.to_array(), colored_ranges.second.to_array())
    )

    combined_stacks: np.ndarray = np.vstack((stacked_colors, stacked_ranges))

    return Image.from_array(combined_stacks, pixel_format=PixelFormat.RGB)



"""Module for time functionality."""

import arrow


def get_time_string(format_string: str = "YYYYMMDD_HHmmss_SSS") -> str:
    """Return the current time as a formatted string."""
    return arrow.get().format(format_string)



"""Unit tests for the image package."""

import pytest
import numpy as np

from mynd.image import Image


@pytest.fixture
def sample_image_data():
    return np.random.randint(0, 256, size=(100, 100, 3), dtype=np.uint8)


@pytest.fixture
def sample_image(sample_image_data):
    return Image.from_array(sample_image_data, Image.Format.RGB)


def test_pixel_format():
    assert Image.Format.GRAY.value == "gray"
    assert Image.Format.RGB.value == "rgb"
    assert Image.Format.RGBA.value == "rgba"


def test_image_layout():
    layout = Image.Layout(height=100, width=200, channels=3)
    assert layout.height == 100
    assert layout.width == 200
    assert layout.channels == 3
    assert layout.shape == (100, 200, 3)


def test_image_properties(sample_image):
    assert sample_image.pixel_format == Image.Format.RGB
    assert sample_image.height == 100
    assert sample_image.width == 100
    assert sample_image.channels == 3
    assert sample_image.shape == (100, 100, 3)
    assert sample_image.dtype == np.uint8
    assert sample_image.ndim == 3


def test_image_from_array(sample_image_data):
    image = Image.from_array(sample_image_data, Image.Format.RGB)
    assert isinstance(image, Image)
    assert image.pixel_format == Image.Format.RGB
    assert image.shape == sample_image_data.shape


def test_image_to_array(sample_image, sample_image_data):
    array = sample_image.to_array()
    assert np.array_equal(array, sample_image_data)
    assert array is not sample_image._data  # Check if it's a copy


def test_image_copy(sample_image):
    copied_image = sample_image.copy()
    assert copied_image is not sample_image
    assert np.array_equal(copied_image.to_array(), sample_image.to_array())


@pytest.mark.parametrize("pixel_format", list(Image.Format))
def test_all_pixel_formats(pixel_format):
    data = np.random.randint(0, 256, size=(10, 10, 3), dtype=np.uint8)
    image = Image.from_array(data, pixel_format)
    assert image.pixel_format == pixel_format


def test_invalid_image_data():
    with pytest.raises(ValueError):
        Image.from_array(
            np.zeros((10, 10, 10, 10)), Image.Format.RGB
        )  # 4D array for RGB



"""Package for schema models."""

from .schemas import ChunkGroupSchema as ChunkGroupSchema
from .schemas import ChunkSchema as ChunkSchema
from .schemas import CalibrationSchema as CalibrationSchema
from .schemas import SensorSchema as SensorSchema
from .schemas import RectifiedSensorSchema as RectifiedSensorSchema
from .schemas import ReferenceSchema as ReferenceSchema
from .schemas import CameraSchema as CameraSchema
from .schemas import PixelMapSchema as PixelMapSchema
from .schemas import StereoSensorPairSchema as StereoSensorPairSchema
from .schemas import StereoCameraPairSchema as StereoCameraPairSchema
from .schemas import StereoPixelMapSchema as StereoPixelMapSchema
from .schemas import StereoRigBase as StereoRigBase
from .schemas import StereoRigSchema as StereoRigSchema
from .schemas import StereoRigWithMapsSchema as StereoRigWithMapsSchema

__all__ = []



"""Module for reading and writing configurations to file. Supported formats are
JSON, YAML, TOML and MSGPACK."""

from pathlib import Path

import msgspec


def read_config(path: Path, mode: str = "r") -> dict | str:
    """Reads a configuration from file. Supported formats are JSON, TOML, YAML,
    and MSGPACK."""
    path: Path = Path(path)

    match path.suffix:
        case ".json":
            return _read_config_json(path, mode)
        case ".toml":
            return _read_config_toml(path, mode)
        case ".yml" | ".yaml":
            return _read_config_yaml(path, mode)
        case ".msgpack":
            return _read_config_msgpack(path, mode)
        case _:
            return f"invalid configuration file format: {path.suffix}"


def _read_config_json(path: Path | str, mode: str = "r") -> dict | str:
    """Reads a configuration from a JSON file."""
    try:
        with open(str(path), mode=mode) as handle:
            data: dict = msgspec.json.decode(handle.read())
            return data
    except IOError as error:
        return str(error)


def _read_config_yaml(path: Path | str, mode: str = "r") -> dict | str:
    """Reads a configuration from a YAML file."""
    try:
        with open(str(path), mode=mode) as handle:
            data: dict = msgspec.yaml.decode(handle.read())
            return data
    except IOError as error:
        return str(error)


def _read_config_toml(path: Path | str, mode: str = "r") -> dict | str:
    """Reads a configuration from a TOML file."""
    try:
        with open(str(path), mode=mode) as handle:
            data: dict = msgspec.toml.decode(handle.read())
            return data
    except IOError as error:
        return str(error)


def _read_config_msgpack(path: Path | str, mode: str = "r") -> dict | str:
    """Reads a configuration from a MSGPACK file."""
    try:
        with open(str(path), mode=mode) as handle:
            data: dict = msgspec.msgpack.decode(handle.read())
            return data
    except IOError as error:
        return str(error)


def write_config(path: Path | str, data: dict, mode: str = "wb") -> Path | str:
    """Writes a configuration to file. Supported formats are JSON, TOML, YAML,
    and MSGPACK."""

    path: Path = Path(path)

    match path.suffix:
        case ".json":
            return _write_config_json(path, data, mode)
        case ".toml":
            return _write_config_toml(path, data, mode)
        case ".yml" | ".yaml":
            return _write_config_yaml(path, data, mode)
        case ".msgpack":
            return _write_config_msgpack(path, data, mode)
        case _:
            return f"invalid configuration file format: {path.suffix}"


def _write_config_json(path: Path | str, data: dict, mode: str = "wb") -> Path | str:
    """Writes a configuration to a JSON file."""
    try:
        with open(str(path), mode) as handle:
            handle.write(msgspec.json.encode(data))
            return path
    except IOError as error:
        return str(error)


def _write_config_yaml(path: Path | str, data: dict, mode: str = "wb") -> Path | str:
    """Writes a configuration to a YAML file."""
    try:
        with open(str(path), mode) as handle:
            handle.write(msgspec.yaml.encode(data))
            return path
    except IOError as error:
        return str(error)


def _write_config_toml(path: Path | str, data: dict, mode: str = "wb") -> Path | str:
    """Writes a configuration to a TOML file."""
    try:
        with open(str(path), mode) as handle:
            handle.write(msgspec.toml.encode(data))
            return path
    except IOError as error:
        return str(error)


def _write_config_msgpack(path: Path | str, data: dict, mode: str = "wb") -> Path | str:
    """Writes a configuration to a MSGPACK file."""
    try:
        with open(str(path), mode) as handle:
            handle.write(msgspec.msgpack.encode(data))
            return path
    except IOError as error:
        return str(error)



"""Module for ASCII key codes."""

from enum import IntEnum


class KeyCode(IntEnum):
    """Class representing ASCII key codes."""

    NULL = 0  # NUL
    BACKSPACE = 8  # BS
    HORIZONTAL_TAB = 9  # HT

    ESC = 27  # ESC
    SPACE = 32  #
    EXCLAMATION = 33  # !
    QUOTE = 34  # "
    NUMBER = 35  # #
    DOLLAR = 36  # $
    PERCENT = 37  # %
    AMPERSAND = 38  # &
    APOSTROPHE = 39  # '
    LEFT_PARANTHESIS = 40  # (
    RIGHT_PARANTHESIS = 41  # )
    ASTERISK = 42  # *
    PLUS = 43  # +
    COMMA = 44  # ,
    HYPHEN = 45  # -
    PERIOD = 46  # .
    SLASH = 47  # /
    D0 = 48  # 0
    D1 = 49  # 1
    D2 = 50  # 2
    D3 = 51  # 3
    D4 = 52  # 4
    D5 = 53  # 5
    D6 = 54  # 6
    D7 = 55  # 7
    D8 = 56  # 8
    D9 = 57  # 9
    COLON = 58  # :
    SEMICOLON = 59  # ;
    LESS_THAN = 60  # <
    EQUAL = 61  # =
    GREATER_THAN = 62  # >
    QUESTION_MARK = 63  # ?
    AT = 64  # @

    UPPERCASE_A = 65
    UPPERCASE_B = 66
    UPPERCASE_C = 67
    UPPERCASE_D = 68
    UPPERCASE_E = 69
    UPPERCASE_F = 70
    UPPERCASE_G = 71
    UPPERCASE_H = 72
    UPPERCASE_I = 73
    UPPERCASE_J = 74
    UPPERCASE_K = 75
    UPPERCASE_L = 76
    UPPERCASE_M = 77
    UPPERCASE_N = 78
    UPPERCASE_O = 79
    UPPERCASE_P = 80
    UPPERCASE_Q = 81
    UPPERCASE_R = 82
    UPPERCASE_S = 83
    UPPERCASE_T = 84
    UPPERCASE_U = 85
    UPPERCASE_V = 86
    UPPERCASE_W = 87
    UPPERCASE_X = 88
    UPPERCASE_Y = 89
    UPPERCASE_Z = 90

    LEFT_BRACKET = 91  # [
    BACKSLASH = 92  # \
    RIGHT_BRACKET = 93  # ]
    HAT = 94  # ^
    UNDERSCORE = 95  # _
    GRAVE_ACCENT = 96  # `

    LEFT_BRACE = 123  # {
    VERTICAL_BAR = 124  # |
    RIGHT_BRACE = 125  # }
    TILDE = 126  # ~
    DEL = 127  # DEL

    LOWERCASE_A = 97
    LOWERCASE_B = 98
    LOWERCASE_C = 99
    LOWERCASE_D = 100
    LOWERCASE_E = 101
    LOWERCASE_F = 102
    LOWERCASE_G = 103
    LOWERCASE_H = 104
    LOWERCASE_I = 105
    LOWERCASE_J = 106
    LOWERCASE_K = 107
    LOWERCASE_L = 108
    LOWERCASE_M = 109
    LOWERCASE_N = 110
    LOWERCASE_O = 111
    LOWERCASE_P = 112
    LOWERCASE_Q = 113
    LOWERCASE_R = 114
    LOWERCASE_S = 115
    LOWERCASE_T = 116
    LOWERCASE_U = 117
    LOWERCASE_V = 118
    LOWERCASE_W = 119
    LOWERCASE_X = 120
    LOWERCASE_Y = 121
    LOWERCASE_Z = 122



"""Package with functionality for registering point clouds."""

from .batch import (
    RegistrationBatch,
    register_batch,
    generate_indices_one_way,
    generate_indices_cascade,
)

from .feature_registrators import (
    extract_fpfh_features,
    register_features_fast,
    register_features_ransac,
)

from .full_registrators import (
    build_pose_graph,
    optimize_pose_graph,
)

from .icp_registrators import (
    register_regular_icp,
    register_colored_icp,
)

from .pipeline import (
    RegistrationPipeline,
    apply_registration_pipeline,
)

from .pipeline_builder import build_registration_pipeline

from .types import (
    Feature,
    RigidTransformation,
    RegistrationResult,
    FeatureExtractor,
    FeatureMatcher,
    PointCloudAligner,
    PointCloudRefiner,
)

from .utilities import log_registration_result


__all__ = [
    "RegistrationBatch",
    "register_batch",
    "generate_indices_one_way",
    "generate_indices_cascade",
    # ...
    "extract_fpfh_features",
    "register_features_fast",
    "register_features_ransac",
    # ...
    "build_pose_graph",
    "optimize_pose_graph",
    # ...
    "register_regular_icp",
    "register_colored_icp",
    # ...
    "RegistrationPipeline",
    "RegistrationCallback",
    "apply_registration_pipeline",
    # ...
    "build_point_cloud_processor",
    "build_ransac_registrator",
    "build_regular_icp_registrator",
    "build_colored_icp_registrator",
    "build_registration_pipeline",
    # ...
    "Feature",
    "RigidTransformation",
    "RegistrationResult",
    "FeatureExtractor",
    "FeatureMatcher",
    "PointCloudAligner",
    "PointCloudRefiner",
    # ...
    "log_registration_result",
]



"""Module with stereo matching interfaces."""

from typing import Callable

from mynd.image import Image
from mynd.utils.containers import Pair


StereoMatcher = Callable[[Image, Image], Pair[Image]]



"""Module for various filesystem functionality, such as filesystem search."""

import glob
import os

from pathlib import Path


def list_directory(directory: Path, pattern: str = "*") -> list[Path]:
    """List files in a directory."""
    return [Path(path) for path in glob.glob(f"{directory}/{pattern}")]


def walk_directory(root: Path) -> list[Path]:
    """List files by walking from a root directory."""
    walker = os.walk(str(root))

    accumulated: list[Path] = list()

    dirpath: str
    dirnames: list[str]
    filenames: list[str]
    for dirpath, dirnames, filenames in walker:
        # TODO: Terminate early at a given depth

        abs_dirnames: list[Path] = [
            Path(f"{dirpath}/{dirname}") for dirname in dirnames
        ]
        abs_filenames: list[Path] = [
            Path(f"{dirpath}/{filename}") for filename in filenames
        ]

        accumulated.extend(abs_dirnames)
        accumulated.extend(abs_filenames)

    return accumulated


def search_files(pattern: str) -> list[Path]:
    """Searches for files with the given pattern."""
    return [Path(path) for path in glob.glob(pattern)]



"""Unit tests for mynds filesystem functionality."""

from pathlib import Path

import pytest
from unittest.mock import patch

from mynd.utils.filesystem import list_directory, search_files


# Tests for list_directory
def test_list_directory():
    with patch("glob.glob") as mock_glob:
        mock_glob.return_value = ["/path/file1.txt", "/path/file2.txt"]
        result = list_directory(Path("/path"), "*.txt")
        assert len(result) == 2
        assert all(isinstance(item, Path) for item in result)
        mock_glob.assert_called_once_with("/path/*.txt")


# Tests for search_files
def test_search_files():
    with patch("glob.glob") as mock_glob:
        mock_glob.return_value = ["/path/file1.txt", "/path/file2.txt"]
        result = search_files("*.txt")
        assert len(result) == 2
        assert all(isinstance(item, Path) for item in result)
        mock_glob.assert_called_once_with("*.txt")



"""Module for common image processors."""

import cv2
import numpy as np

from .image_types import Image, PixelFormat


def flip_image(image: Image, axis: int = 1) -> Image:
    """Flip an image around the specified axis."""
    flipped_values: np.ndarray = cv2.flip(image.to_array(), axis)
    flipped: Image = Image.from_array(
        data=flipped_values,
        pixel_format=image.pixel_format,
    )
    return flipped


def resize_image(image: Image, height: int, width: int) -> Image:
    """Resize an image to the desired size. The size is specified as HxW"""
    return Image.from_array(
        data=cv2.resize(image.to_array(), (width, height), cv2.INTER_AREA),
        pixel_format=image.pixel_format,
    )


def process_image_clahe(image: Image, size: int, clip: float) -> Image:
    """Filter an image with CLAHE."""
    # NOTE: This function only works on uint8 images for now
    assert image.dtype == np.uint8, f"invalid image dtype: {image.dtype}"

    clahe = cv2.createCLAHE(clipLimit=clip, tileGridSize=(size, size))

    match image.pixel_format:
        case PixelFormat.GRAY:
            return _process_image_clahe_gray(image, clahe)
        case PixelFormat.RGB | PixelFormat.BGR:
            return _process_image_clahe_rgb(image, clahe)
        case _:
            raise NotImplementedError(f"invalid pixel format: {image.pixel_format}")

    raise NotImplementedError


def _process_image_clahe_rgb(image: Image, processor: object) -> Image:
    """Applies CLAHE to a RGB / BGR image via LAB color space."""

    match image.pixel_format:
        case PixelFormat.RGB:
            format_to = cv2.COLOR_RGB2LAB
            format_from = cv2.COLOR_LAB2RGB
        case PixelFormat.BGR:
            format_to = cv2.COLOR_BGR2LAB
            format_from = cv2.COLOR_LAB2BGR
        case _:
            raise NotImplementedError

    values: np.ndarray = cv2.cvtColor(image.to_array(), format_to)
    values[:, :, 0]: np.ndarray = processor.apply(values[:, :, 0])
    values: np.ndarray = cv2.cvtColor(values, format_from)

    return Image.from_array(values, image.pixel_format)


def _process_image_clahe_gray(image: Image, processor: object) -> Image:
    """Applies CLAHE to a grayscale image."""
    values: np.ndarray = image.to_array()
    values: np.ndarray = processor.apply(values)
    return Image.from_array(values, image.pixel_format)


def process_image_gamma(image: Image, gamma: float) -> Image:
    """Corrects the colors of an image with gamma correction."""
    values: np.ndarray = image.to_array()
    min_value: int | float = np.iinfo(image.dtype).min
    max_value: int | float = np.iinfo(image.dtype).max

    scale: int | float = max_value - min_value
    difference: np.ndarray = values - min_value

    corrected_values: np.ndarray = np.array(
        scale * (difference / scale) ** gamma, dtype=image.dtype
    )

    return Image.from_array(corrected_values, pixel_format=image.pixel_format)


def process_image_linear(image: Image, brightness: float, contrast: float) -> Image:
    """Corrects an image by apply a linear scale to the image brightness and contrast.
    :arg brightness:    scale value for the image brightness
    :arg contrast:      scale value for the image contrast
    """
    values: np.ndarray = image.to_array()

    min_value: int | float = np.iinfo(image.dtype).min
    max_value: int | float = np.iinfo(image.dtype).max
    scale: int | float = max_value - min_value

    normalized_values: np.ndarray = np.array(
        ((values - min_value) * brightness) / scale * contrast + 0.5 * (1 - contrast),
        dtype=np.float32,
    )

    # Clamp values that are outside the interval
    normalized_values[normalized_values > 1.0] = 1.0
    normalized_values[normalized_values < 0] = 0.0

    # Rescale values back to the range of the image
    corrected_values: np.ndarray = np.array(
        scale * normalized_values + min_value,
        dtype=image.dtype,
    )

    return Image.from_array(corrected_values, image.pixel_format)


def normalize_image(
    image: Image,
    lower: int | float | None = None,
    upper: int | float | None = None,
    flip: bool = False,
) -> Image:
    """Normalizes the values of an image to be between the lower and upper values."""
    values: np.ndarray = image.to_array()

    if lower:
        values[values < lower] = lower
    if upper:
        values[values > upper] = upper

    min_value: int | float = values.min()
    max_value: int | float = values.max()

    # TODO: Add support for multiple dtypes here
    if flip:
        scale: int = -255
        offset: int = 255
    else:
        scale: int = 255
        offset: int = 0

    # Normalized values between 0 and 1
    normalized: np.ndarray = (values - min_value) / (max_value - min_value)
    normalized: np.ndarray = scale * normalized + offset

    normalized: np.ndarray = normalized.astype(np.uint8)
    return Image.from_array(normalized, image.pixel_format)



"""Module for point cloud types."""

from collections.abc import Callable
from typing import TypeAlias

import open3d


PointCloud: TypeAlias = open3d.geometry.PointCloud
PointCloudLoader: TypeAlias = Callable[[None], PointCloud | str]
PointCloudProcessor = Callable[[PointCloud], PointCloud]



"""Package for database functionality."""

from .common import Engine as Engine
from .common import Session as Session
from .common import create_engine as create_engine
from .common import verify_engine as verify_engine

__all__ = []



"""Module for IO functionality for images."""

from pathlib import Path

import imageio.v3 as iio
import numpy as np

from .image_types import Image, PixelFormat


def _infer_pixel_format(values: np.ndarray) -> PixelFormat:
    """Infers the image format based on the image data type and channel count."""

    dtype: np.dtype = values.dtype
    channels: int = values.shape[2]

    is_floating_point: bool = dtype in [np.float16, np.float32, np.float64]

    match channels:
        case 1:
            return PixelFormat.X if is_floating_point else PixelFormat.GRAY
        case 3:
            return PixelFormat.XYZ if is_floating_point else PixelFormat.RGB
        case 4:
            return PixelFormat.XYZW if is_floating_point else PixelFormat.RGBA
        case _:
            raise ValueError(
                f"invalid combination of dtype and channels: {dtype}, {values}"
            )


def read_image(
    uri: str | Path,
    *,
    index: int | None = None,
    plugin: str | None = None,
    extension: str | None = None,
    **kwargs,
) -> Image | str:
    """Reads an image from a uniform resource identifier (URI). The image format must
    either be GRAY, X, RGB, XYZ, RGBA, or XYZW. Returns an image with dimensions HxWxC.
    """
    try:
        values: np.ndarray = iio.imread(
            uri, index=index, plugin=plugin, extension=extension, **kwargs
        )

        if values.ndim != 3:
            values: np.ndarray = np.expand_dims(values, axis=2)

        _metadata: dict = iio.immeta(uri)
        pixel_format: PixelFormat = _infer_pixel_format(values)
        return Image.from_array(data=values, pixel_format=pixel_format)
    except (OSError, IOError, TypeError, ValueError) as error:
        return str(error)


def write_image(
    image: Image | np.ndarray,
    uri: str | Path,
    *,
    plugin: str | None = None,
    extension: str | None = None,
    **kwargs,
) -> Path | str:
    """Writes an image to a uniform resource identifier (URI). The image format
    must either be GRAY, RGB, or RGBA."""

    if isinstance(image, Image):
        values: np.ndarray = image.to_array()
    elif isinstance(image, np.ndarray):
        values: np.ndarray = image
    else:
        return f"invalid image type: {type(image)}"

    try:
        iio.imwrite(
            uri,
            np.squeeze(values),
            plugin=None,
            extension=None,
            format_hint=None,
            **kwargs,
        )
        return uri
    except (OSError, IOError, TypeError, ValueError) as error:
        return str(error)



"""Module to perform assimilation of camera references."""

import numpy as np
import polars as pl

from scipy.spatial.transform import Rotation

import mynd.schemas as schemas

from mynd.utils.log import logger

from .alignment import (
    RigidTransformation,
    align_points_arun,
    align_points_umeyama,
)
from .transforms import (
    convert_locations_geodetic2enu,
    convert_locations_enu2geodetic,
    convert_locations_geodetic2ned,
    convert_locations_ned2geodetic,
)


def assimilate_camera_references(
    cameras: list[schemas.CameraSchema],
) -> list[schemas.CameraSchema]:
    """Performs an assimilation between prior and aligned camera references."""

    prior_references: pl.DataFrame = _tabulate_camera_references_prior(cameras)
    aligned_references: pl.DataFrame = _tabulate_camera_references_aligned(cameras)

    # Assimilate camera references by using the
    assimilated_references: pl.DataFrame = _assimilate_camera_references_full_impl(
        prior_references,
        aligned_references,
    )

    references: dict[int, schemas.CameraSchema.ReferenceType] = (
        _parse_camera_references(assimilated_references)
    )

    assimilated_cameras: list[schemas.CameraSchema] = list()
    for camera in cameras:
        reference: schemas.CameraSchema.ReferenceType | None = references.get(camera.id)

        if reference is None:
            continue

        camera.assimilated_reference = reference
        assimilated_cameras.append(camera)

    return assimilated_cameras


def _parse_camera_references(
    table: pl.DataFrame,
) -> dict[int, schemas.CameraSchema.ReferenceType]:
    """Parse table rows as camera references."""
    references: dict[int, schemas.CameraSchema.ReferenceType] = dict()
    for row in table.iter_rows(named=True):
        camera_id: int | None = row.get("camera_id")

        if camera_id is None:
            continue

        reference: schemas.CameraSchema.ReferenceType = (
            schemas.CameraSchema.ReferenceType(
                epsg_code=4326,
                longitude=row.get("longitude"),
                latitude=row.get("latitude"),
                height=row.get("height"),
                yaw=row.get("yaw"),
                pitch=row.get("pitch"),
                roll=row.get("roll"),
            )
        )

        references[camera_id] = reference

    return references


def _tabulate_camera_references_prior(
    cameras: list[schemas.CameraSchema],
) -> pl.DataFrame:
    """Tabulates prior camera references."""
    return pl.DataFrame(
        [
            {
                "camera_id": camera.id,
                "camera_label": camera.label,
                "longitude": camera.prior_reference.longitude,
                "latitude": camera.prior_reference.latitude,
                "height": camera.prior_reference.height,
                "yaw": camera.prior_reference.yaw,
                "pitch": camera.prior_reference.pitch,
                "roll": camera.prior_reference.roll,
            }
            for camera in cameras
            if camera.prior_reference
        ]
    )


def _tabulate_camera_references_aligned(
    cameras: list[schemas.CameraSchema],
) -> pl.DataFrame:
    """Tabulates aligned camera references."""
    return pl.DataFrame(
        [
            {
                "camera_id": camera.id,
                "camera_label": camera.label,
                "longitude": camera.aligned_reference.longitude,
                "latitude": camera.aligned_reference.latitude,
                "height": camera.aligned_reference.height,
                "yaw": camera.aligned_reference.yaw,
                "pitch": camera.aligned_reference.pitch,
                "roll": camera.aligned_reference.roll,
            }
            for camera in cameras
            if camera.aligned_reference
        ]
    )


def _assimilate_camera_references_full_impl(
    reference_priors: pl.DataFrame,
    reference_estimates: pl.DataFrame,
) -> pl.DataFrame:
    """Assimilate reference priors and estimates. The assimilation process
    aligns prior locations to estimates, and uses the aligned priors to fill in
    missing location estimates."""

    GEODESIC_COLUMNS: list[str] = ["latitude", "longitude", "height"]
    CARTESIAN_COLUMNS: list[str] = ["north", "east", "down"]

    for column in GEODESIC_COLUMNS:
        assert column in reference_priors.columns, f"missing prior column: {column}"
        assert (
            column in reference_estimates.columns
        ), f"missing estimate column: {column}"

    # Get centre latitude and longitude
    center_latitude: float = reference_priors.select(pl.median("latitude")).item()
    center_longitude: float = reference_priors.select(pl.median("longitude")).item()
    center_height: float = 0.0

    # Remove old reference information
    reference_priors_base: pl.DataFrame = reference_priors.select(
        pl.exclude(["latitude", "longitude", "height", "yaw", "pitch", "roll"])
    )

    # Convert reference priors to NED
    reference_priors_ned: np.ndarray = convert_locations_geodetic2ned(
        reference_priors.select(GEODESIC_COLUMNS).to_numpy(),
        center_longitude=center_longitude,
        center_latitude=center_latitude,
        center_height=center_height,
    )

    # Convert reference estimates to ENU
    reference_estimates_ned: np.ndarray = convert_locations_geodetic2ned(
        reference_estimates.select(GEODESIC_COLUMNS).to_numpy(),
        center_longitude=center_longitude,
        center_latitude=center_latitude,
        center_height=center_height,
    )

    reference_priors_ned: pl.DataFrame = pl.from_numpy(
        reference_priors_ned, schema=CARTESIAN_COLUMNS
    )

    reference_estimates_ned: pl.DataFrame = pl.from_numpy(
        reference_estimates_ned, schema=CARTESIAN_COLUMNS
    )

    # Add cartesian coordinates to prior references as columns
    reference_priors: pl.DataFrame = pl.concat(
        [reference_priors, reference_priors_ned], how="horizontal"
    )
    reference_estimates: pl.DataFrame = pl.concat(
        [reference_estimates, reference_estimates_ned], how="horizontal"
    )

    matched_priors, matched_estimates = _match_references_by_column(
        reference_priors,
        reference_estimates,
        column="camera_id",
    )

    transformation: RigidTransformation = align_points_umeyama(
        matched_priors.select(CARTESIAN_COLUMNS).to_numpy(),
        matched_estimates.select(CARTESIAN_COLUMNS).to_numpy(),
    )

    # Correct prior locations
    updated_location_priors_ned: np.ndarray = _transform_locations(
        reference_priors.select(CARTESIAN_COLUMNS).to_numpy(),
        transformation,
    )

    # Correct prior attitudes
    updated_attitude_priors: np.ndarray = _transform_attitudes(
        reference_priors.select(["yaw", "pitch", "roll"]).to_numpy(),
        transformation,
    )

    # Convert updated locations priors from ENU to WGS84
    updated_location_priors_geo: np.ndarray = convert_locations_ned2geodetic(
        updated_location_priors_ned,
        center_longitude=center_longitude,
        center_latitude=center_latitude,
        center_height=center_height,
    )

    # Combine attributes, udpated locations, and update attitudes
    updated_reference_priors: pl.DataFrame = pl.concat(
        [
            reference_priors_base,
            pl.from_numpy(
                updated_location_priors_geo, schema=["latitude", "longitude", "height"]
            ),
            pl.from_numpy(updated_attitude_priors, schema=["yaw", "pitch", "roll"]),
        ],
        how="horizontal",
    )

    # Overwrite updated priors with estimates to obtain the assimilated references
    assimilated_references: pl.DataFrame = updated_reference_priors.update(
        reference_estimates, on="camera_id"
    )

    return assimilated_references


def _match_references_by_column(
    reference_priors: pl.DataFrame,
    reference_estimates: pl.DataFrame,
    column: str,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    """TODO"""

    assert (
        column in reference_priors.columns
    ), f"missing column in reference priors: {column}"
    assert (
        column in reference_estimates.columns
    ), f"missing column in reference estimates: {column}"

    # Find common camera labels between prior and aligned references
    matching_labels: pl.DataFrame = (
        reference_priors.join(
            reference_estimates,
            on=column,
            how="inner",
        )
        .select(column)
        .unique()
    )

    # Select prior and estimated references from the matching labels
    matched_priors: pl.DataFrame = reference_priors.filter(
        pl.col(column).is_in(matching_labels.get_column(column))
    )
    matched_estimates: pl.DataFrame = reference_estimates.filter(
        pl.col(column).is_in(matching_labels.get_column(column))
    )

    return matched_priors, matched_estimates


def _transform_locations(
    locations: np.ndarray, transformation: RigidTransformation
) -> np.ndarray:
    """Transforms locations by applying a rigid-body transformation."""
    transformed_locations: np.ndarray = transformation.apply(locations)
    return transformed_locations


def _transform_attitudes(
    attitudes: np.ndarray, transformation: RigidTransformation
) -> np.ndarray:
    """Transforms attitudes by applying a rigid-body transformation."""
    common_rotation: Rotation = Rotation.from_matrix(transformation.rotation)
    attitudes: Rotation = Rotation.from_euler(seq="ZYX", angles=attitudes, degrees=True)

    # Apply the common rotation to the attitudes
    transformed_attitudes: np.ndarray = (common_rotation * attitudes).as_euler(
        seq="ZYX", degrees=True
    )

    # NOTE: Correct yaw angle to interval [0, 360], as Scipy uses [-180, 180]
    mask: np.ndarray = transformed_attitudes[:, 0] < 0.0
    transformed_attitudes[mask, 0] = 360.0 + transformed_attitudes[mask, 0]

    return transformed_attitudes
